from random import randint
import statsmodels.api as sm
import numpy as np
import math
from misc import compute_moments, plot_histogram, print_moments
from input_output import model_io

'''
This script is used for bootstrapping the distributions of the regression coefficients.
'''

num_iterations = 10000
show_results = True


def print_progress(iteration):
    progress = (iteration+1)/num_iterations
    end = " " if progress < 1 else "\n"
    print("\rBootstrap progress {:2.2%}".format(progress), end=end)


def _compute_moments_and_corr(matrix):
    moments = compute_moments(matrix, 4)
    corr = np.corrcoef(matrix)
    for i in range(len(corr)):
        corr[i] = [0. if math.isnan(x) else x for x in corr[i]]
        corr[i][i] = 1.
    return np.array(moments), np.array(corr)


def case_resampling(data_set, use_export, use_multiphase, m_estimator):
    x, y = data_set.get_all_data(use_export, use_multiphase)
    n, m = len(x), len(x[0])
    regression_coefficients = [[] for _ in range(m)]
    for j in range(num_iterations):
        print_progress(j)
        sample_x, sample_y = [], []
        for _ in range(n):
            i = randint(0, n-1)
            sample_x.append(x[i])
            sample_y.append(y[i])
        results = sm.RLM(sample_y, sample_x, M=m_estimator()).fit()
        for i in range(m):
            regression_coefficients[i].append(results.params[i])
    moments, corr = _compute_moments_and_corr(regression_coefficients)
    model_io.save_coefficients(regression_coefficients, data_set.name)
    if show_results:
        print_moments(moments, data_set.num_wells)
        plot_histogram(regression_coefficients, moments, data_set.num_wells)
    return moments, corr


def residuals_resampling(data_set, use_export, use_multiphase, m_estimator):
    x, y = data_set.get_all_data(use_export, use_multiphase)
    n, m = len(x), len(x[0])
    results = sm.OLS(y, x).fit()
    y_fitted = results.predict(x)
    residuals = [y[i] - y_fitted[i] for i in range(n)]
    regression_coefficients = [[] for _ in range(m)]
    for j in range(num_iterations):
        print_progress(j)
        sample_y = [y_fitted[i] + residuals[randint(0, n-1)] for i in range(n)]
        results = sm.RLM(sample_y, x, M=m_estimator()).fit()
        for i in range(m):
            regression_coefficients[i].append(results.params[i])
    model_io.save_coefficients(regression_coefficients, data_set.name)
    moments, corr = _compute_moments_and_corr(regression_coefficients)
    if show_results:
        print_moments(moments, data_set.num_wells)
        plot_histogram(regression_coefficients, moments, data_set.num_wells)
    return moments, corr


def point_resampling(data_set, m_estimator):
    n, m = data_set.num_points, data_set.num_wells + 1
    regression_coefficients = [[] for _ in range(m)]
    for j in range(num_iterations):
        print_progress(j)
        x, y = data_set.get_data_sample()
        results = sm.RLM(y, x, M=m_estimator()).fit()
        for i in range(m):
            regression_coefficients[i].append(results.params[i])
    moments, corr = _compute_moments_and_corr(regression_coefficients)
    model_io.save_coefficients(regression_coefficients, data_set.name)
    if show_results:
        print_moments(moments, data_set.num_wells)
        plot_histogram(regression_coefficients, moments, data_set.num_wells)
    return moments, corr
