from misc import get_limit, plot_histogram
from scenario_generation import moment_matching
import numpy as np
from collections import OrderedDict
import itertools
from perfect_info_solver import PISolver


def generate_mosel_matrix(name, matrix):
    s = name+" : [\n"
    for row in matrix:
        s += ' '.join([str(x) for x in row])+"\n"
    s += "]"
    return s


# Compute upper/lower limit of each choke.
def chokes_limit(chokes, lower):
    limits = []
    for i in range(chokes.shape[1]):
        limits.append(get_limit(np.sort(chokes[:, i]), lower))
        f = min if lower else max
        limits[-1] = f(limits[-1], chokes[-1][i])
    return np.array(limits)


# Create a "deterministic" ordering of the chokes.
def create_priority_list(oil_contributions, gas_contributions):
    pri_list = []
    large_constant = 10**10
    for oil, gas in zip(oil_contributions, gas_contributions):
        if oil == 0:
            oil = 10 ** (-10)
        gor = gas/oil
        if oil < 0:
            gor = 1/gor if gor > 0 else gor + large_constant
        pri_list.append(gor)
    return np.array(pri_list).argsort() + 1


# Find all the unique "modes" in the data set.
# Return only modes that are represented by at least <min_count> rows, that have no more than <max_turned_off> chokes
# turned off.
def get_modes(chokes):
    min_count = 5
    max_turned_off = 2
    modes = {}
    for i in range(chokes.shape[0]):
        key = tuple(int(v > 0.01) for v in chokes[i])
        if len(key) - np.sum(key) <= max_turned_off:
            modes.setdefault(key, []).append(i)
    return OrderedDict({key: modes[key] for key in modes if len(modes[key]) > min_count})


# Generate all possible "modes" that have at most <max_turned_off> chokes turned off
def get_all_modes(num_chokes):
    max_turned_off = 2
    modes = []
    for i in range(max_turned_off+1):
        for bits in itertools.combinations(range(num_chokes), i):
            s = [1] * num_chokes
            for bit in bits:
                s[bit] = 0
            modes.append(tuple(s))
    return modes


# Used to fit mode constants when multiphase constants are not available.
class MoselInputGenerator:
    def __init__(self, chokes, gas_export, oil_export, oil_moments, oil_corr, gas_moments, gas_corr):
        self.oil_moments = oil_moments
        self.oil_corr = oil_corr
        self.gas_moments = gas_moments
        self.gas_corr = gas_corr

        self.chokes = chokes
        self.oil_export = oil_export
        self.gas_export = gas_export
        self.num_wells = chokes.shape[1]
        self.num_constants = oil_moments.shape[1] - self.num_wells
        self.modes = get_modes(chokes) if self.num_constants == 1 else get_all_modes(self.num_wells)

        self.chokes_lower_limit = chokes_limit(chokes, lower=True)
        self.chokes_upper_limit = chokes_limit(chokes, lower=False)
        self.chokes_current = chokes[-1]
        self.mode_current = self._get_current_mode()

        # Deterministic parameters
        self.oil_det_contributions = oil_moments[0][self.num_constants:]
        self.gas_det_contributions = gas_moments[0][self.num_constants:]
        self.oil_det_constants = self._constants(oil_moments[0], np.expand_dims(self.oil_det_contributions, 0),
                                                 oil_export)
        self.gas_det_constants = self._constants(gas_moments[0], np.expand_dims(self.gas_det_contributions, 0),
                                                 gas_export)

        self.scenario_generator = moment_matching.ScenarioGenerator
        self.test_scenario_generator = moment_matching.ScenarioGenerator

        self.oil_test_scenarios = None
        self.gas_test_scenarios = None
        self.pi_solver = None
        self.pi_value = 0

    def _compute_bigm(self, a, b, delta_gas, dim=2):
        return np.sum(a * (self.chokes_upper_limit - self.chokes_current), axis=dim-1) + b.max(axis=dim-1) - delta_gas

    def _get_current_mode(self):
        key_curr = tuple(int(v > 0.01) for v in self.chokes_current)
        for key, i in zip(self.modes, range(len(self.modes))):
            if key == key_curr:
                return i
        return -1

    def _constants(self, scenarios, contributions, export, delta=True):
        if self.num_constants == 1:
            constants = []
            for c in contributions:
                constants.append([])
                for m in self.modes:
                    i = self.modes[m]
                    constants[-1].append(np.mean(export[i] - np.sum(c*self.chokes[i], axis=1)))
            constants = np.array(constants)
        else:
            constants = np.dot(scenarios[:self.num_constants].T, np.array(self.modes).T)
        if constants.ndim == 1:
            return constants - constants[self.mode_current] if delta else constants
        else:
            return constants - np.expand_dims(constants[:, self.mode_current], 1) if delta else constants

    def _contributions(self, scenarios):
        return scenarios[self.num_constants:].T

    def run(self, num_points, num_scenarios, num_test_scenarios, delta_gas, penalty, use_prev_test=False):
        try:
            oil_scenarios = self.scenario_generator(self.oil_moments, self.oil_corr, num_scenarios, "oil").run()
            gas_scenarios = self.scenario_generator(self.gas_moments, self.gas_corr, num_scenarios, "gas").run()
        except RuntimeError:
            return False

        # When doing i.e. stability testing we want test scenarios to stay the same for each test with the same setup
        if not use_prev_test:
            try:
                self.oil_test_scenarios = self.test_scenario_generator(self.oil_moments, self.oil_corr,
                                                                       num_test_scenarios, "oil (test)").run()
                self.gas_test_scenarios = self.test_scenario_generator(self.gas_moments, self.gas_corr,
                                                                       num_test_scenarios, "gas (test)").run()
                self.pi_solver = PISolver(self.oil_test_scenarios, self.gas_test_scenarios, self.chokes_current,
                                          self.chokes_upper_limit, self.chokes_lower_limit, num_points-1,
                                          delta_gas)
                self.pi_value = self.pi_solver.run()
            except RuntimeError:
                return False

        oil_contributions = self._contributions(oil_scenarios)
        gas_contributions = self._contributions(gas_scenarios)
        oil_constants = self._constants(oil_scenarios, oil_contributions, self.oil_export)
        gas_constants = self._constants(gas_scenarios, gas_contributions, self.gas_export)

        oil_test_contributions = self._contributions(self.oil_test_scenarios)
        gas_test_contributions = self._contributions(self.gas_test_scenarios)
        oil_test_constants = self._constants(self.oil_test_scenarios, oil_test_contributions, self.oil_export)
        gas_test_constants = self._constants(self.gas_test_scenarios, gas_test_contributions, self.gas_export)

        pri_list = create_priority_list(self.oil_det_contributions, self.gas_det_contributions)

        det_bigm = self._compute_bigm(self.gas_det_contributions, self.gas_det_constants, delta_gas, dim=1)
        scen_bigm = self._compute_bigm(gas_contributions, gas_constants, delta_gas)
        test_bigm = self._compute_bigm(gas_test_contributions, gas_test_constants, delta_gas)

        text = "numChokes : {0}\n\n".format(self.num_wells)
        text += "numScenarios : {0}\n\n".format(num_scenarios)
        text += "numTestScenarios : {0}\n\n".format(num_test_scenarios)
        text += "numPoints : {0}\n\n".format(num_points)
        text += "deltaGas : {0}\n\n".format(delta_gas)
        text += "downChangePenalty : {0}\n\n".format(penalty)
        text += "currChokeOpening : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.chokes_current))
        text += "chokesUpperLimit : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.chokes_upper_limit))
        text += "chokesLowerLimit : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.chokes_lower_limit))
        text += "priorityList : [\n{0}\n]\n\n".format(' '.join(str(i) for i in pri_list))
        text += "expProdImprovementPI : {0}\n\n".format(self.pi_value)

        text += "numModes : {0}\n\n".format(len(self.modes))
        text += generate_mosel_matrix("modes", list(self.modes.keys()) if self.num_constants == 1 else self.modes)+"\n\n"
        text += "D_oilContribution : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.oil_det_contributions))
        text += "D_gasContribution : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.gas_det_contributions))
        text += generate_mosel_matrix("oilContribution", oil_contributions) + "\n\n"
        text += generate_mosel_matrix("gasContribution", gas_contributions) + "\n\n"
        text += generate_mosel_matrix("T_oilContribution", oil_test_contributions)+"\n\n"
        text += generate_mosel_matrix("T_gasContribution", gas_test_contributions)+"\n\n"

        text += "D_oilConstant : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.oil_det_constants))
        text += "D_gasConstant : [\n{0}\n]\n\n".format(' '.join(str(i) for i in self.gas_det_constants))
        text += generate_mosel_matrix("oilConstant", oil_constants)+"\n\n"
        text += generate_mosel_matrix("gasConstant", gas_constants)+"\n\n"
        text += generate_mosel_matrix("T_oilConstant", oil_test_constants)+"\n\n"
        text += generate_mosel_matrix("T_gasConstant", gas_test_constants)+"\n\n"

        text += "D_M : {0}\n\n".format(det_bigm)
        text += "M : [\n{0}\n]\n\n".format(' '.join(str(i) for i in scen_bigm))
        text += "T_M : [\n{0}\n]\n\n".format(' '.join(str(i) for i in test_bigm))

        with open("mosel/"+"problemData.txt", "w") as f:
            f.write(text)
        return True

    def plot_distributions(self, name=None, skip_constants=False):
        oil_name = name+"_oil" if name is not None else None
        gas_name = name+"_gas" if name is not None else None
        if skip_constants:
            plot_histogram(self.oil_test_scenarios[self.num_constants:], self.oil_moments[:, self.num_constants:],
                           self.num_wells, _include_time=False, title="Oil", save_as=oil_name)
            plot_histogram(self.gas_test_scenarios[self.num_constants:], self.gas_moments[:, self.num_constants:],
                           self.num_wells, _include_time=False,  title="Gas", save_as=gas_name)
        else:
            plot_histogram(self.oil_test_scenarios, self.oil_moments, self.num_wells, _include_time=False, title="Oil",
                           save_as=oil_name)
            plot_histogram(self.gas_test_scenarios, self.gas_moments, self.num_wells, _include_time=False,  title="Gas",
                           save_as=gas_name)

