import matplotlib.pyplot as plt
import numpy as np
from math import ceil
from scipy.stats import norm
from my_collections.petroleum_data import include_time
import csv

result_folder_path = 'test_results/'

'''
This is a collection of methods that are used by several of the modules.
'''


# Find the upper or lower limit of a choke opening.
# This is the region in which you "trust" the regression results.
def get_limit(x, minimum=True):
    max_dist = 0.01
    min_density = 5
    density = 0
    indices = range(1, x.size) if minimum else range(x.size-1, 0, -1)
    for i in indices:
        if x[i] < 0.01:
            continue
        if x[i] - x[i-1] < max_dist:
            density += 1
        else:
            density = 0
        if density == min_density:
            break
    return x[i-min_density] if minimum else x[i+min_density-1]


def plot_histogram(value_arr, moments, num_wells, title=None, _include_time=None, save_as=None):
    if _include_time is None:
        _include_time = include_time

    columns = 5
    rows = ceil(len(value_arr)/float(columns))
    num_var = len(value_arr)

    plt.figure(facecolor="white")
    for i in range(num_var):
        row = i//columns
        column = i % columns
        ax = plt.subplot2grid((rows, columns), (row, column))

        if i == num_var - 1 and _include_time:
            title = "Time"
        elif i > num_var - num_wells - (2 if _include_time else 1):
            title = "Slope well {}".format(i - (num_var - num_wells) + (2 if _include_time else 1))
        else:
            title = "Constant well {}".format(i+1)
        ax.set_title(title)
        values = value_arr[i]
        num_bins = 50
        ax.hist(values, bins=num_bins, normed=True, facecolor='green', alpha=0.75)
        std4 = 4*moments[1, i]**0.5
        x = np.linspace(moments[0][i] - std4, moments[0][i] + std4) if std4 != 0 else [moments[0, i]]
        plt.xticks([moments[0][i] - 3*std4/4, moments[0][i], moments[0][i] + 3*std4/4], fontsize=16)
        if moments[1, i] != 0:
            ax.plot(x, norm.pdf(x, moments[0, i], moments[1, i]**0.5), 'r-', linewidth=2)
        ax.grid(True)
    if title:
        plt.title(title)
    plt.gcf().set_size_inches(3*columns+6, 3*rows, forward=True)
    if save_as is not None:
        plt.savefig(result_folder_path+save_as)
    plt.show()


def plot_histogram2(moments):
    num_var = moments.shape[1]
    plt.figure(facecolor="white")
    axes = [plt.subplot2grid((2, 1), (0, 0)), plt.subplot2grid((2, 1), (1, 0))]
    lines, labels = [], []
    for i in range(num_var):
        row = i // (num_var//2)
        ax = axes[row]
        if i == 0:
            ax.set_title("Constants")
        if i == num_var//2:
            ax.set_title("Slopes")
        std4 = 4*moments[1, i]**0.5
        x = np.linspace(moments[0][i] - std4, moments[0][i] + std4) if std4 != 0 else [moments[0, i]]
        y = norm.pdf(x, moments[0, i], moments[1, i]**0.5)
        line, = ax.plot(x, y, linewidth=2)
        ax.fill_between(x, y, 0, facecolor=line.get_color(), alpha=0.5)
        if row:
            lines.append(line)
            labels.append("Well {}".format(i % (num_var//2)+1))
    plt.figlegend(lines, labels, loc="lower right", fancybox=True, framealpha=0.75)
    plt.tight_layout()
    plt.show()


def plot_branch_results():
    with open("test_results/branch.csv") as csv_file:
        reader = csv.reader(csv_file, delimiter=";")
        next(reader)
        time = []
        branch = []
        no_branch = []
        i = 0
        for row in reader:
            time.append((i+1)*30)
            branch.append(float(row[1]))
            no_branch.append(float(row[2]))
            i += 1
    plt.figure(facecolor="white")
    plt.xlim([30, 300])
    plt.plot(time, branch, label="With branching strategy")
    plt.plot(time, no_branch, label="Without branching strategy")
    plt.xlabel("Time [s]")
    plt.ylabel("Gap")
    plt.grid()
    plt.legend()
    plt.show()


def print_moments(moments, num_wells, _include_time=None):
    if _include_time is None:
        _include_time = include_time
    num_var = moments.shape[1]
    print("{:20}".format("") + "\t".join("{:>10}".format("Moment "+str(i)) for i in range(1, 5)))
    for i in range(num_var):
        if i == num_var - 1 and _include_time:
            label = "Time"
        elif i > num_var - num_wells - (2 if _include_time else 1):
            label = "Slope well {}".format(i - (num_var - num_wells) + (2 if _include_time else 1))
        else:
            label = "Constant well {}".format(i+1)
        print("{:20}".format(label) + "\t".join("{:10.2f}".format(x) if abs(x) < 1e6 else "{:10.2e}".format(x)
                                                for x in moments[:, i]))


def print_correlations(corr, num_wells, _include_time=None):
    if _include_time is None:
        _include_time = include_time
    num_var = corr.shape[0]
    labels = []
    for i in range(num_var):
        if i == num_var - 1 and _include_time:
            labels.append("T")
        elif i > num_var - num_wells - (2 if _include_time else 1):
            labels.append("A{}".format(i - (num_var - num_wells) + (2 if _include_time else 1)))
        else:
            labels.append("B{}".format(i+1))
    print("{:5}".format("") + "".join("{:>10}".format(l) for l in labels))
    for i in range(num_var):
        print("{:5}".format(labels[i])+"".join("{:10.2f}".format(corr[i][j]) for j in range(num_var)))


# a is a single variable (i.e. set of samples)
def _moment(a, i):
    mean = np.mean(a)
    return mean if i == 1 else np.mean((a - mean) ** i)


# a is a matrix with variables as rows.
# Returns a matrix with one row for each moment, and one column for each variable.
def compute_moments(a, n):
    return np.array([[_moment(row, i) for row in a] for i in range(1, n+1)])


# Moments is a matrix with one row for each moment, and one column for each variable.
def normalize_moments(moments):
    np.seterr(invalid='ignore')
    num_variables = moments.shape[1]
    num_mom = moments.shape[0]
    std = moments[1]**0.5
    higher_mom = [moments[i]/std**(i+1) for i in range(2, num_mom)]
    return np.nan_to_num(np.array([[0.] * num_variables, [1.] * num_variables, *higher_mom]))
