import matplotlib.pyplot as plt
import numpy as np
from builtins import print

from input_output import data_io
import statsmodels.api as sm
from math import ceil
from sklearn.cross_validation import KFold
from misc import get_limit


# Parameters for this script
use_export = True
use_multiphase = True


class RegressionModel:
    def __init__(self, name, M):
        self.name = name
        self.M = M
        self.model = None

        # Used if a separate set of constants must be fitted for plotting of the multiphase models
        self.constants = None

    def fit(self, x, y):
        self.model = sm.RLM(y, x, M=self.M()).fit()

    @property
    def params(self):
        if self.model is None:
            return None
        if self.constants:
            return self.constants + list(self.model.params[1:])
        return self.model.params

    def predict(self, x):
        if self.model is None:
            return [0]*len(x)
        return self.model.predict(x)

    def summary(self):
        return self.model.summary2()


def plot_export_models():
    plt.figure(facecolor="white")
    x, y = data.get_test_data(range(data.num_points), use_multiphase)
    time = np.arange(0, data.num_points)
    plt.plot(time, y,  linewidth=2, label="Real", alpha=0.5)
    for m in models:
        plt.plot(time, m.predict(x), label=m.name)
    plt.ylabel(data.name.capitalize()+" [Sm^3/h]")
    plt.xlabel("Time")
    plt.xticks([])
    plt.legend(loc=4)
    plt.grid()
    fig = plt.gcf()
    fig.set_size_inches(10, 6.2, forward=True)
    plt.show()


def _get_multiphase_data(i):
    x, y = np.array(data.chokes)[:, i], np.array(data.multiphase)[:, i]
    to_delete = [i for i in range(data.num_points) if x[i] < 0.01 or y[i] < 0.01]
    x, y = np.delete(x, to_delete), np.delete(y, to_delete)
    p = x.argsort()
    return x[p], y[p]


def plot_multiphase_models():
    plt.figure(facecolor="white")
    columns = 4
    rows = ceil(data.num_wells / float(columns))
    for i in range(data.num_wells):
        row = i//columns
        column = i % columns
        ax = plt.subplot2grid((rows, columns), (row, column))
        x, y = _get_multiphase_data(i)
        ax.plot(x, y, 'b.', label="Real")
        for m in models:
            a = m.params[i + data.num_wells]
            b = m.params[i]
            ax.plot(x, a*x + b, label=m.name)
        ax.set_title("Well {}".format(i+1))
        ax.grid(True)
        ax.axvline(get_limit(x, minimum=True), ls='--', color='r')
        ax.axvline(get_limit(x, minimum=False), ls='--', color='r')
        leg = ax.legend(loc=2, fancybox=True, fontsize=10)
        leg.get_frame().set_alpha(0.5)
    plt.show()


def create_models():
    _models = []
    x, y = data.get_all_data(use_export, use_multiphase)
    for name, M in rlm_funcs:
        _models.append(RegressionModel(name, M))
        _models[-1].fit(x, y)
    return _models


# Used to fit individual constants for each well, when only export measurements have been used to fit the actual models
# These constants are only used for plotting purposes.
def fit_model_constants():
    constants = [[] for _ in models]
    for i in range(data.num_wells):
        x = data.chokes[:, i]
        y = data.multiphase[:, i]
        for j in range(len(models)):
            constants[j].append(np.mean(y - models[j].params[i + 1] * x))
    for j in range(len(models)):
        models[j]._constants = constants[j]


# Root mean squared error
def rmse(real, predicted):
    return np.mean((predicted-real)**2)**0.5


# Mean absolute deviation
def mad(real, predicted):
    return np.mean(abs(predicted-real))


def compute_accuracy(real, predicted):
    print("\n{:15} {:>15} {:>15}".format("", "RMSE", "MAD"))
    for i in range(num_rlm_funcs):
        print("{:<15} {:>15.2f} {:>15.2f}".format(rlm_funcs[i][0], rmse(real, predicted[i]), mad(real, predicted[i])))


def r_squared(model):
    x, y = data.get_all_data(use_export, use_multiphase)
    print("R^2 of RLM-fit with M={}: {}".format(model.name, np.corrcoef(model.predict(x), y)[0, 1]**2))


def r_squared_per_well(model):
    print("R^2 of RLM with M={}".format(model.name))
    for i in range(data.num_wells):
        x, y = _get_multiphase_data(i)
        y_pred = model.params[i] + model.params[i+data.num_wells]*x
        print("Well {}: {}".format(i+1, np.corrcoef(y, y_pred)[0, 1]**2))
    x, y = data.get_test_data(range(data.num_points), use_multiphase)
    print("Export: {}".format(np.corrcoef(y, model.predict(x))[0, 1]**2))


# Runs 10-fold cross validation on the data.
# Returns a set of predicted values for each model, and a set of the actual values.
def create_and_validate_models():
    data.shuffle()
    num_folds, fold_num = 10, 0
    kf = KFold(data.num_points, n_folds=num_folds)
    y_predicted, y_real = [[] for _ in rlm_funcs], []
    print("Running {}-fold cross-validation.".format(num_folds))
    for train_indices, test_indices in kf:
        x_train, y_train = data.get_training_data(train_indices, use_export, use_multiphase)
        x_test, y_test = data.get_test_data(test_indices, use_multiphase)
        for i in range(num_rlm_funcs):
            model = sm.RLM(y_train, x_train, M=rlm_funcs[i][1]()).fit()
            y_predicted[i].extend(model.predict(x_test))
            print("\rValidation progress {:2.2%}".format((fold_num*num_rlm_funcs+i)/(num_rlm_funcs*num_folds)), end="")
        y_real.extend(y_test)
        fold_num += 1
    print("\rValidation progress {:2.2%}".format(1), end="\n")

    compute_accuracy(np.array(y_real), np.array(y_predicted))
    data.sort()


rlm_funcs = [("LeastSquares", sm.robust.norms.LeastSquares),
             ("AndrewWave", sm.robust.norms.AndrewWave),
             ("HuberT", sm.robust.norms.HuberT),
             ("Hampel", sm.robust.norms.Hampel),
             ("RamsayE", sm.robust.norms.RamsayE),
             ("TukeyBiweight", sm.robust.norms.TukeyBiweight)]
num_rlm_funcs = len(rlm_funcs)


oil_data, gas_data = data_io.read_raw_data()
data = gas_data
data.compress()
# create_and_validate_models()
models = create_models()
r_squared_per_well(models[2])
plot_export_models()
if not use_multiphase:
    fit_model_constants()
plot_multiphase_models()
