import numpy as np


w_export = 1/0.001
w_multi = 1/0.05
include_time = True
merge_threshold = 1e-2


def distance(a1, a2):
    return (np.sum(a1 - a2) ** 2) ** 0.5


# Data set of "5 minute samples". Contains both export data and multiphase data.
class PetroleumData:
    def __init__(self, chokes, export, multiphase, name):
        self.chokes = np.array(chokes)
        self.export = np.array(export)
        self.multiphase = np.array(multiphase)
        self.name = name

        self.num_wells = len(chokes[0])
        self.num_points = len(chokes)
        self.time = np.arange(self.num_points)

    def get_training_data(self, indices, use_export, use_multiphase):
        x, y = [], []
        if use_multiphase:
            for i in indices:
                for j in range(self.num_wells):
                    if self.chokes[i][j] < 0.01 or self.multiphase[i][j] < 0.01:
                        continue
                    multi_row = [0]*2*self.num_wells
                    multi_row[j] = 1
                    multi_row[self.num_wells + j] = self.chokes[i][j]
                    if include_time:
                        multi_row.append(self.time[i]/self.num_wells)
                    y.append(self.multiphase[i][j])
                    x.append(np.array(multi_row))
                if use_export:
                    export_row = [1 if self.chokes[i][j] > 0.01 else 0 for j in range(self.num_wells)]
                    export_row.extend(self.chokes[i])
                    if include_time:
                        export_row.append(self.time[i])
                    y.append(self.export[i])
                    x.append(np.array(export_row))
        elif use_export:
            for i in indices:
                y.append(self.export[i])
                x.append([1.] + [v for v in self.chokes[i]])
                if include_time:
                    x[-1].append(self.time[i])
        return np.array(x), np.array(y)

    def get_test_data(self, indices, use_multiphase):
        x, y = [], []
        for i in indices:
            if use_multiphase:
                row = [1 if self.chokes[i][j] > 0.01 else 0 for j in range(self.num_wells)]
            else:
                row = [1]
            row.extend(self.chokes[i])
            if include_time:
                row.append(self.time[i])
            y.append(self.export[i])
            x.append(np.array(row))
        return np.array(x), np.array(y)

    def get_all_data(self, use_export, use_multiphase):
        return self.get_training_data(range(self.num_points), use_export, use_multiphase)

    # Used when "include_time" is true. This compression scheme maintains the time aspect.
    def _get_merge_intervals(self):
        merge = [[0]]
        j = 0
        for i in range(1, self.num_points):
            print("\rCompression progress {:2.2%}".format(j/self.num_points), end="")
            if distance(self.chokes[j], self.chokes[i]) <= merge_threshold:
                merge[-1].append(i)
            else:
                j = i
                merge.append([j])
        if j < i:
            merge.append(list(range(j, i+1)))
        print("\rCompression progress {:2.2%}".format(1))
        return merge

    # Used when "include_time" is false. This compression scheme does not maintain the time aspect.
    def _get_merge_intervals2(self):
        merge = []
        indices = list(range(self.num_points))
        while indices:
            print("\rCompression progress {:2.2%}".format((self.num_points-len(indices))/self.num_points), end="")
            new_indices = []
            merge.append([indices[0]])
            for i in range(1, len(indices)):
                if distance(self.chokes[indices[0]], self.chokes[indices[i]]) <= merge_threshold:
                    merge[-1].append(indices[i])
                else:
                    new_indices.append(indices[i])
            indices = new_indices
        print("\rCompression progress {:2.2%}".format(1))
        return merge

    def compress(self):
        print("Compressing {} samples.".format(self.num_points))
        if include_time:
            merge_intervals = self._get_merge_intervals()
        else:
            merge_intervals = self._get_merge_intervals2()
        self.chokes = np.array([np.mean(self.chokes[p], axis=0) for p in merge_intervals])
        self.export = np.array([np.mean(self.export[p]) for p in merge_intervals])
        self.multiphase = np.array([np.mean(self.multiphase[p], axis=0) for p in merge_intervals])
        self.time = np.array([np.mean(self.time[p]) for p in merge_intervals])
        self.num_points = len(self.chokes)
        print("Compressed to {} samples.".format(self.num_points))

    # p is an array of indices used to order the data of this set
    def _order(self, p):
        self.chokes = self.chokes[p]
        self.export = self.export[p]
        self.multiphase = self.multiphase[p]
        self.time = self.time[p]

    def shuffle(self):
        self._order(np.random.permutation(self.num_points))
        print("Data set shuffled.")

    def sort(self):
        self._order(self.time.argsort())
        print("Data set sorted")


# Data set of "aggregated data points" (~5 hour samples). This data set contains export data only.
class AggregatedPetroleumData(PetroleumData):
    def __init__(self, chokes, export, chokes_var, export_var, name):
        PetroleumData.__init__(self, chokes, export, None, name)
        self.chokes_std = np.array([[x**0.5 for x in row] for row in chokes_var])
        self.export_std = np.array([x**0.5 for x in export_var])

    def get_training_data(self, indices):
        return super(AggregatedPetroleumData, self).get_training_data(indices, True, False)

    def get_test_data(self, indices, use_multiphase):
        return super(AggregatedPetroleumData, self).get_test_data(indices, False)

    def get_all_data(self):
        return super(AggregatedPetroleumData, self).get_training_data(range(self.num_points), True, False)

    '''
    Each data point in the aggregated data set is uncertain, with given mean and std.
    This information can be used to draw samples for each point. The set of samples will give a "unique" model, and
    by building a model several times a distribution for each model coefficient can be determined.
    '''
    def get_data_sample(self):
        x, y = [], []
        for i in range(self.num_points):
            x.append([1.])
            for j in range(self.num_wells):
                if self.chokes_std[i][j] > 0:
                    x[-1].append(np.random.normal(self.chokes[i][j], self.chokes_std[i][j]))
                else:
                    x[-1].append(self.chokes[i][j])
            y.append(np.random.normal(self.export[i], self.export_std[i]))
        return np.array(x), np.array(y)