import numpy as np


class PISolver:
    def __init__(self, oil_scenarios, gas_scenarios, chokes, chokes_u, chokes_l, num_steps, delta_gas):
        self.num_wells = oil_scenarios.shape[0] // 2
        self.num_scenarios = oil_scenarios.shape[1]
        self.num_steps = num_steps

        self.oil_b = oil_scenarios[:self.num_wells].T
        self.gas_b = gas_scenarios[:self.num_wells].T
        self.oil_a = oil_scenarios[self.num_wells:].T
        self.gas_a = gas_scenarios[self.num_wells:].T

        self.goal_delta_gas = delta_gas
        self.chokes = chokes
        self.chokes_l = chokes_l
        self.chokes_u = chokes_u
        self.num_off = np.size(self.chokes) - np.count_nonzero(self.chokes)
        self._is_feasible = None

    def _solve_scenario(self, s, chokes, step, num_steps, num_off, delta_gas, obj_val):
        if step == num_steps:
            if delta_gas <= self.goal_delta_gas:
                return obj_val
            return -float('inf')
        options = [obj_val] if delta_gas <= self.goal_delta_gas else []
        for i in chokes:
            d1 = self.chokes_u[i] - self.chokes[i]
            d2 = self.chokes_l[i] - self.chokes[i]
            d3 = -self.chokes[i]

            # Stop before upper limit
            if delta_gas + self.gas_a[s][i]*d1 > self.goal_delta_gas > delta_gas:
                options.append(obj_val + self.oil_a[s][i]*(self.goal_delta_gas-delta_gas)/self.gas_a[s][i])

            # Adjust to upper limit
            options.append(self._solve_scenario(s, chokes[chokes != i], step+1, num_steps, num_off,
                                                delta_gas + self.gas_a[s][i]*d1, obj_val + self.oil_a[s][i]*d1))

            # Stop before upper limit
            if delta_gas + self.gas_a[s][i]*d2 < self.goal_delta_gas < delta_gas:
                options.append(obj_val + self.oil_a[s][i]*(self.goal_delta_gas-delta_gas)/self.gas_a[s][i])

            # Adjust to lower limit
            options.append(self._solve_scenario(s, chokes[chokes != i], step+1, num_steps, num_off,
                                                delta_gas + self.gas_a[s][i]*d2, obj_val + self.oil_a[s][i]*d2))

            # Turn off well
            if num_off < 2:
                options.append(self._solve_scenario(s, chokes[chokes != i], step+1, num_steps, num_off+1,
                                                    delta_gas + self.gas_a[s][i]*d3 - self.gas_b[s][i],
                                                    obj_val + self.oil_a[s][i]*d3 - self.oil_b[s][i]))
        return max(options)

    def run(self, verbose=True):
        value = 0
        for s in range(self.num_scenarios):
            if verbose:
                print("\rPI solver progress: {}%".format((s+1)*100/self.num_scenarios), end="")
            value += self._solve_scenario(s, np.arange(self.num_wells), 0, self.num_steps, self.num_off, 0, 0)
        if verbose:
            print()
        self._is_feasible = value > -float('inf')
        return value/self.num_scenarios

    def is_feasible(self):
        if self._is_feasible is None:
            self.run(verbose=False)
        return self._is_feasible


