import numpy as np
import scipy.optimize
import warnings
from misc import compute_moments, print_moments

warnings.filterwarnings('ignore', 'The iteration is not making good progress')


# Used for memoization of function calls to "mom"
mom_memo = []


# Moments of the original variable
def mom(a, i):
    if i > len(mom_memo)-1:
        mom_memo.append(np.mean(a**i))
    return mom_memo[i]


# Moments of the transformed variable
# Col "0" is the base variant used for computing the obj_val
# The other cols are used for the gradient and hessian
mom_new = np.array([])


# compute the "transformed" moment n, given the parameters p, and base array _x
def mom_f(p, _x, n):
    c = p
    for i in range(n-1):
        c = np.convolve(p, c)
    num_mom = mom_new.shape[0] - 1
    for i in range(min(2, num_mom-n)*(num_mom-1)+1):
        mom_new[n][i] = sum(c[j] * mom(_x, j+i) for j in range(len(c)))
    return mom_new[n][0]


# the gradient of obj_val with respect to the parameters
# j = the variable the derivative is taken with respect to
# k = a particular moment
def obj_grad(p, _x, i, target_mom):
    num_mom = target_mom.shape[0]
    prime = np.empty(num_mom, dtype=np.float64)
    for j in range(num_mom):
        for k in range(num_mom):
            prime[j] += 2*(k+1)*mom_new[k][j]*(mom_new[k+1][0] - target_mom[k][i])
    return prime


# the hessian of obj_val with respect to the parameters
# j = the variable the first derivative is taken with respect to
# k = the variable the second derivative is taken with respect to
# l = a particular moment
def obj_hess(p, _x, i, target_mom):
    num_mom = target_mom.shape[0]
    hess = np.empty((num_mom, num_mom), dtype=np.float64)
    for l in range(num_mom):
        for j in range(l, num_mom):
            hess[l][j] = 2*mom_new[0][l]*mom_new[0][j]
            for k in range(1, num_mom):
                hess[l][j] += 2*(k+1)*((k+1)*mom_new[k][l]*mom_new[k][j] +
                                       k*mom_new[k-1][l+j]*(mom_new[k+1][0] - target_mom[k][i]))
            hess[j][l] = hess[l][j]
    return hess


def obj_val(p, _x, i, target_mom):
    return sum((mom_f(p, _x[i], j+1)-target_mom[j][i])**2 for j in range(target_mom.shape[0]))


def _distance_mom(m1, m2):
    return np.mean((m1 - m2)**2)**0.5


def reset_mom(_x, num_mom):
    global mom_memo, mom_new
    mom_new = np.empty((num_mom+1, 2*num_mom-1), dtype=np.float64)
    for i in range(2*num_mom-1):
        mom_new[0][i] = mom(_x, i)
    mom_memo = []


def transform(_x, target_mom):
    num_mom = target_mom.shape[0]
    epsilon_x = 0.2 if _x.shape[1] > 5 else 0.1
    max_iter = 3
    error1 = _distance_mom(compute_moments(_x, num_mom), target_mom)
    j = 0
    while error1 > epsilon_x and j < max_iter:
        for i in range(len(_x)):
            reset_mom(_x[i], num_mom)
            p0 = [0., 1.] + [0.]*(num_mom-2)
            p = scipy.optimize.fmin_ncg(obj_val, p0, fprime=obj_grad, fhess=obj_hess, args=(_x, i, target_mom), disp=0)
            _x[i] = sum(p[j]*_x[i]**j for j in range(num_mom))
        error2 = _distance_mom(compute_moments(_x, num_mom), target_mom)
        j = j + 1 if error1 - error2 < 1e-4 else 0
        error1 = error2
    return _x
