import numpy as np
import scenario_generation.cubic_transformation as cubic
from misc import normalize_moments


class ScenarioGenerator:
    def __init__(self, target_mom, target_corr, num_scenarios, name=None):
        # Number of variables
        self.num_variables = len(target_mom[0])

        # Number of scenarios
        self.num_scenarios = num_scenarios

        # Target moments
        self.target_mom = target_mom

        # Target correlations
        self.target_corr = target_corr

        # Cholesky decomposition of target correlation
        self.L = np.linalg.cholesky(target_corr)

        # Identity matrix
        self.I = np.identity(self.num_variables, dtype=np.float64)

        # Target standard deviation
        self.target_std = target_mom[1] ** 0.5

        # Target variance
        self.target_mean = target_mom[0]

        self.name = name

        self.epsilon_y = 0.02 if self.num_variables > 5 else 0.01

        self.z = []

    def run(self):
        if np.count_nonzero(self.target_mom[1]) == 0:
            self.z = np.array([[self.target_mom[0][i]]*self.num_scenarios for i in range(self.num_variables)])
        else:
            norm_moments = normalize_moments(self.target_mom)
            not_converged = True
            while not_converged:
                x = self.draw_scenarios(norm_moments)
                try:
                    y, not_converged = self.fix_correlations(x, norm_moments)
                except np.linalg.linalg.LinAlgError:
                    raise RuntimeError("Failed generating scenarios.")
            self.z = self.back_transform(y)
        message = "Generated {} scenarios".format(self.num_scenarios)
        print(message + ("." if self.name is None else " for {}.".format(self.name)))
        return self.z

    def draw_scenarios(self, _moments):
        scenarios = []
        for i in range(self.num_variables):
            scenarios.append(np.random.normal(_moments[0][i], _moments[1][i]**0.5, self.num_scenarios))
        return np.array(scenarios, dtype=np.float64)

    def _distance_corr(self, r1, r2):
        iu = np.triu_indices(self.num_variables, k=1)
        return np.mean((r1[iu]-r2[iu])**2)**0.5

    def fix_correlations(self, _x, _norm_mom):
        y_p = np.dot(self.L, _x)
        r_p = np.corrcoef(y_p)
        i, max_iter = 0, 10
        while 1:
            l_p = np.linalg.cholesky(r_p)
            y_star = np.dot(np.linalg.inv(l_p), y_p)
            y_star2 = np.dot(self.L, y_star)
            y_p = cubic.transform(y_star2, _norm_mom)
            r_p = np.corrcoef(y_p)
            i += 1
            if self._distance_corr(r_p, self.target_corr) < self.epsilon_y or i == max_iter:
                break
        return y_p, i == max_iter

    def back_transform(self, _y):
        return np.expand_dims(self.target_std, 1) * _y + np.expand_dims(self.target_mean, 1)

