import os
import numpy as np
import csv
from input_output import mosel_io
import matplotlib.pyplot as plt
from scenario_generation import moment_matching, random_sampling


result_folder_path = 'test_results/'


# Generates a set of synthetic models, i.e. moments and correlations for n variables
# Only works when the model has been built on multiphase data
def _generate_synthetic_models(real_moments, num_wells, num_synthetic_wells):
    nc = real_moments.shape[1] - num_wells
    constants_mean_range = np.min(real_moments[0, :nc]), np.max(real_moments[0, :nc])
    slopes_mean_range = np.min(real_moments[0, nc:]), np.max(real_moments[0, nc:])

    relative_var = np.abs(real_moments[1]/real_moments[0])
    constants_var_range = np.min(relative_var[:nc]), np.max(relative_var[:nc])
    slopes_var_range = np.min(relative_var[nc:]), np.max(relative_var[nc:])
    _synthetic_moments = []

    # Mean
    c = np.random.uniform(constants_mean_range[0], constants_mean_range[1], num_synthetic_wells)
    s = np.random.uniform(slopes_mean_range[0], slopes_mean_range[1], num_synthetic_wells)
    _synthetic_moments.append(np.append(c, s))

    # Variance
    c = np.random.uniform(constants_var_range[0], constants_var_range[1], num_synthetic_wells)*np.abs(c)
    s = np.random.uniform(slopes_var_range[0], slopes_var_range[1], num_synthetic_wells)*np.abs(s)
    _synthetic_moments.append(np.append(c, s))

    # Skew and kurtosis
    _synthetic_moments.append([0]*num_synthetic_wells*2)
    _synthetic_moments.append(np.ones(num_synthetic_wells*2)*3 * _synthetic_moments[1] ** 2)
    return np.array(_synthetic_moments), np.identity(2*num_synthetic_wells)


# ----- BRANCHING STRATEGY -----

def plot_branching_strategy():
    with open("test_results/branch.csv") as csv_file:
        data_reader = csv.reader(csv_file, delimiter=";")
        t = []
        gap1, gap2 = [], []
        next(data_reader)
        for row in data_reader:
            t.append(int(row[0]))
            gap1.append(float(row[1]))
            gap2.append(float(row[2]))
    plt.figure(facecolor="white")
    plt.grid()
    plt.plot(t, gap1, label="Customized Branching Strategy")
    plt.plot(t, gap2, label="Default Branching Strategy")
    plt.legend()
    plt.xlabel("Time [s]")
    plt.ylabel("Optimality Gap [Fractional]")
    plt.xlim(xmin=min(t))
    plt.savefig(result_folder_path + "branching")


# ----- RELAXATION TESTING -----


def test_relaxation(mosel_ig, num_scenarios):
    num_tests = 10
    for i in range(num_tests):
        print("Running relaxation test {} of {}.".format(i+1, num_tests))
        mosel_ig.run(num_points=4, num_scenarios=num_scenarios, num_test_scenarios=10000, delta_gas=0, penalty=10)
        os.system("mosel exec mosel/strategy RELAXED=false INFILE='mosel/problemData.txt' "
                  "OUTFILE='mosel/{}_scenarios.csv'".format(num_scenarios))
        os.system("mosel exec mosel/strategy RELAXED=true INFILE='mosel/problemData.txt' "
                  "OUTFILE='mosel/results/{}_scenarios_relaxed.csv'".format(num_scenarios))


# ----- CAPACITY TESTING -----


def _parse_capacity_results(name):
    with open("mosel/results/{}__capacity_test_results.csv".format(name)) as csv_file:
        data_reader = csv.reader(csv_file, delimiter=",")
        result = {}
        for row in data_reader:
            wells = int(row[0])
            steps = int(row[1])
            scenarios = int(row[2])
            gap = float(row[6])
            gap = "Inf." if gap > 1e12 else "{:.3}".format(-gap)
            result.setdefault((wells, steps), {})[scenarios] = gap
            next(data_reader)
    with open("{}{}__capacity_test_results.csv".format(result_folder_path, name), 'w') as csv_file:
        data_writer = csv.writer(csv_file, delimiter=",", lineterminator='\n')
        data_writer.writerow(["", "", "Scenarios"])
        scenarios_row = sorted([s for s in result[list(result.keys())[0]]])
        data_writer.writerow(["Wells", "Steps"] + scenarios_row)
        for key in sorted(result):
            wells, steps = key
            data_writer.writerow([wells, steps] + [result[key][s] if s in result[key] else "-" for s in scenarios_row])


def test_problem_capacity__strategy(oil_moments, gas_moments, actual_num_wells):
    test_num_wells = [3, 7, 15, 20]
    test_num_points = [3, 4, 5]
    test_num_scenarios = {3: [20, 50, 100, 200], 7: [20, 50, 100, 200], 15: [50, 100, 200], 20: [50, 100, 200]}

    for w in test_num_wells:
        _oil_moments, _oil_corr = _generate_synthetic_models(oil_moments, actual_num_wells, w)
        _gas_moments, _gas_corr = _generate_synthetic_models(gas_moments, actual_num_wells, w)
        _chokes = np.array([[0.3]*w]*6 + [[0.7]*w]*6 + [[0.5]*w])
        _mosel_ig = mosel_io.MoselInputGenerator(_chokes, None, None, _oil_moments, _oil_corr, _gas_moments, _gas_corr)
        for s in test_num_scenarios[w]:
            for p in test_num_points:
                print("----------------------------------------------------------------------------")
                print("Capacity testing with {} wells, {} scenarios, and {} points.".format(w, s, p))
                _mosel_ig.run(num_points=p, num_scenarios=s, num_test_scenarios=10000, delta_gas=0, penalty=10)
                os.system("mosel exec mosel/strategy_v2 INFILE='mosel/problemData.txt' "
                          "OUTFILE='mosel/results/strategy__capacity_test_results.csv' ONLY_SS=true VERBOSE=false")
    _parse_capacity_results("strategy")


def test_problem_capacity__penalty(oil_moments, gas_moments, actual_num_wells):
    test_num_wells = [3, 7, 15, 20]
    test_num_points = [3, 4, 5]
    test_num_scenarios = [50, 100, 150, 200]

    for w in test_num_wells:
        _oil_moments, _oil_corr = _generate_synthetic_models(oil_moments, actual_num_wells, w)
        _gas_moments, _gas_corr = _generate_synthetic_models(gas_moments, actual_num_wells, w)
        _chokes = np.array([[0.3]*w]*6 + [[0.7]*w]*6 + [[0.5]*w])
        _mosel_ig = mosel_io.MoselInputGenerator(_chokes, None, None, _oil_moments, _oil_corr, _gas_moments, _gas_corr)
        for s in test_num_scenarios:
            for p in test_num_points:
                print("----------------------------------------------------------------------------")
                print("Capacity testing with {} wells, {} scenarios, and {} points.".format(w, s, p))
                _mosel_ig.run(num_points=p, num_scenarios=s, num_test_scenarios=10000, delta_gas=0, penalty=10)
                os.system("mosel exec mosel/penalty INFILE='mosel/problemData.txt' "
                          "OUTFILE='mosel/results/penalty__capacity_test_results.csv' ONLY_SS=true VERBOSE=false")
    _parse_capacity_results("penalty")


# ----- STABILITY TESTING -----


def _plot_std(std_dicts, labels, plot_name):
    plt.figure(facecolor="white")
    for i in range(len(std_dicts)):
        x, y = [], []
        for key in sorted(std_dicts[i]):
            x.append(key)
            y.append(std_dicts[i][key])
        plt.plot(x, y, label=labels[i])
    plt.grid()
    plt.legend()
    plt.xlabel("Number of scenarios")
    plt.ylabel("Standard dev. of objective")
    plt.savefig(result_folder_path + plot_name)


def _parse_stability_results(name):
    with open("mosel/results/{}__stability_test_results.csv".format(name)) as csv_file:
        data_reader = csv.reader(csv_file, delimiter=",")
        in_sample = {}
        out_of_sample = {}
        for row in data_reader:
            num_scen = int(row[2])
            if float(row[3]) <= 599:
                in_sample.setdefault(num_scen, []).append(float(row[5])/num_scen)
                out_of_sample.setdefault(num_scen, []).append(float(row[7])/1000)
            next(data_reader)
        for num_scen in in_sample:
            in_sample[num_scen] = np.std(in_sample[num_scen])
        for num_scen in out_of_sample:
            out_of_sample[num_scen] = np.std(out_of_sample[num_scen])
    return in_sample, out_of_sample


def test_stability__strategy(mosel_ig):
    test_num_scenarios = [20, 50, 70, 100]
    for s in test_num_scenarios:
        for i in range(10):
            use_prev_test = i != 0 or s != test_num_scenarios[0]
            print("----------------------------------------------------------------------------")
            print("Stability testing with {} scenarios ({}).".format(s, i+1))
            mosel_ig.scenario_generator = moment_matching.ScenarioGenerator
            mosel_ig.run(num_points=4, num_scenarios=s, num_test_scenarios=10000, delta_gas=0, penalty=10,
                         use_prev_test=use_prev_test)
            os.system("mosel exec mosel/strategy_v2 INFILE='mosel/problemData.txt' "
                      "OUTFILE='mosel/results/strategy1__stability_test_results.csv' VERBOSE=false")
            mosel_ig.scenario_generator = random_sampling.ScenarioGenerator
            mosel_ig.run(num_points=4, num_scenarios=s, num_test_scenarios=10000, delta_gas=0, penalty=10,
                         use_prev_test=True)
            os.system("mosel exec mosel/strategy_v2 INFILE='mosel/problemData.txt' "
                      "OUTFILE='mosel/results/strategy2__stability_test_results.csv' VERBOSE=false")

    in_sample1, out_of_sample1 = _parse_stability_results("strategy1")
    in_sample2, out_of_sample2 = _parse_stability_results("strategy2")
    _plot_std([in_sample1, in_sample2], ["Moment Matching", "Random Sampling"], "strategy__in_sample_stability")
    _plot_std([out_of_sample1, out_of_sample2], ["Moment Matching", "Random Sampling"], "strategy__out_of_sample_stability")


def test_stability__penalty(mosel_ig):
    test_num_scenarios = [20, 50, 70, 100]
    for s in test_num_scenarios:
        for i in range(10):
            use_prev_test = i != 0 or s != test_num_scenarios[0]
            print("----------------------------------------------------------------------------")
            print("Stability testing with {} scenarios ({}).".format(s, i+1))
            mosel_ig.scenario_generator = moment_matching.ScenarioGenerator
            mosel_ig.run(num_points=4, num_scenarios=s, num_test_scenarios=10000, delta_gas=0, penalty=10,
                         use_prev_test=use_prev_test)
            os.system("mosel exec mosel/penalty INFILE='mosel/problemData.txt' "
                      "OUTFILE='mosel/results/penalty1__stability_test_results.csv' VERBOSE=false")
            mosel_ig.scenario_generator = random_sampling.ScenarioGenerator
            mosel_ig.run(num_points=4, num_scenarios=s, num_test_scenarios=10000, delta_gas=0, penalty=10,
                         use_prev_test=True)
            os.system("mosel exec mosel/penalty INFILE='mosel/problemData.txt' "
                      "OUTFILE='mosel/results/penalty2__stability_test_results.csv' VERBOSE=false")

    in_sample1, out_of_sample1 = _parse_stability_results("penalty1")
    in_sample2, out_of_sample2 = _parse_stability_results("penalty2")
    _plot_std([in_sample1, in_sample2], ["Moment Matching", "Random sampling"], "penalty__in_sample_stability")
    _plot_std([out_of_sample1, out_of_sample2], ["Moment Matching", "Random sampling"], "penalty__out_of_sample_stability")


# ----- ECONOMIC TESTING -----


num_scenarios = 50
num_test_scenarios = 10000


def _parse_real_case_results(name):
    print("Uncertainty in obj. value for {} model:\n".format(name))
    with open("mosel/results/{}__economic_test_results2.csv".format(name)) as csv_file:
        data_reader = csv.reader(csv_file, delimiter=",")
        print("{:<12}{:12}{:<10}{:<10}".format("Num. steps", "Delta gas", "Std (SS)", "Std (EV)"))
        for _, num_steps, delta_gas in data_reader:
            ss_values = np.array([float(f) for f in next(data_reader)])
            ev_values = np.array([float(f) for f in next(data_reader)])
            print("{:<12}{:<12}{:<10.2}{:<10.2}".format(num_steps, delta_gas, np.std(ss_values), np.std(ev_values)))
            plt.figure(facecolor="white")
            plt.rc('text', usetex=True)
            plt.hist(ss_values, bins=50, normed=True, alpha=0.50, label="Stochastic model")
            plt.hist(ev_values, bins=50, normed=True, alpha=0.50, label="Deterministic model")
            plt.title(r"{}, $\Delta$Gas={}, {} Changes".format(name.capitalize(), delta_gas.strip(), num_steps.strip()),
                      fontsize=20)
            plt.tick_params(labelsize=20)
            plt.xlabel("Objective Value",  fontsize=20)
            plt.legend()
            plt.savefig(result_folder_path+"obj_values/{}__{}steps_{}gas".format(name, num_steps, delta_gas))
            plt.clf()
    print("\n")


def test_real_case(mosel_ig):
    mosel_ig.chokes_upper_limit += 0.1
    test_num_points = [3, 4]
    for g in [-50000, 0, 10000]:
        for p in test_num_points:
            print("----------------------------------------------------------------------------")
            while 1:
                mosel_ig.run(num_points=p, num_scenarios=num_scenarios, num_test_scenarios=num_test_scenarios,
                             delta_gas=g, penalty=10)
                if mosel_ig.pi_solver.is_feasible():
                    break
                print("At least one infeasible scenario in test set. Restarting scenario generation.")
            print("Economic testing with delta gas = {} and {} points (strategy).".format(g, p))
            os.system("mosel exec mosel/strategy_v3 INFILE='mosel/problemData.txt' "
                      "OUTFILE='mosel/results/strategy__economic_test_results.csv' "
                      "OUTFILE2='mosel/results/strategy__economic_test_results2.csv' WRITE_ALL_OBJ=true")
            if g < 0:
                continue
            print("----------------------------------------------------------------------------")
            print("Economic testing with delta gas = {} and {} points (penalty).".format(g, p))
            os.system("mosel exec mosel/penalty INFILE='mosel/problemData.txt' "
                      "OUTFILE='mosel/results/penalty__economic_test_results.csv' "
                      "OUTFILE2='mosel/results/penalty__economic_test_results2.csv' WRITE_ALL_OBJ=true")
    _parse_real_case_results("strategy")
    _parse_real_case_results("penalty")
    mosel_ig.chokes_upper_limit -= 0.1


# ----- ECONOMIC TESTING (SYNTHETIC CASES) -----


def _run_synthetic_case(oil_moments, gas_moments, case_num):
    num_wells = oil_moments.shape[1]//2
    corr = np.identity(2*num_wells, dtype=np.float64)
    chokes = np.array([[0.3]*num_wells]*6 + [[0.7]*num_wells]*6 + [[0.5]*num_wells])
    mosel_ig = mosel_io.MoselInputGenerator(chokes, None, None, oil_moments, corr, gas_moments, corr)
    for g in [-50000, 0, 10000]:
        while 1:
            mosel_ig.run(num_points=4, num_scenarios=num_scenarios, num_test_scenarios=num_test_scenarios,
                         delta_gas=g, penalty=10)
            if mosel_ig.pi_solver.is_feasible():
                break
            print("At least one infeasible scenario in test set. Restarting scenario generation.")
        os.system("mosel exec mosel/strategy_v3 INFILE='mosel/problemData.txt' "
                  "OUTFILE='mosel/results/strategy__case{}_test_results.csv'".format(case_num))
        if g < 0:
            continue
        os.system("mosel exec mosel/penalty INFILE='mosel/problemData.txt' "
                  "OUTFILE='mosel/results/penalty__case{}_test_results.csv'".format(case_num))


def test_synthetic_case1():
    num_wells = 5
    mean = np.ones(num_wells)*50
    variance = np.array([4, 6, 8, 10, 12])**2
    skew = np.zeros(num_wells)
    kurt = np.ones(num_wells)*3*variance**2
    oil_moments = np.append(np.zeros((4, num_wells)), np.array([mean, variance, skew, kurt]), axis=1)
    gas_moments = np.zeros((4, 2*num_wells))
    gas_moments[0, num_wells:] = np.ones(num_wells)*50000
    _run_synthetic_case(oil_moments, gas_moments, 1)


def test_synthetic_case2():
    num_wells = 5
    mean = np.ones(num_wells)*50
    variance = np.ones(num_wells)*(5**2)
    skew = np.array([-0.5, -0.3, 0, 0.3, 0.5])*variance**1.5
    kurt = np.ones(num_wells)*3*variance**2
    oil_moments = np.append(np.zeros((4, num_wells)), np.array([mean, variance, skew, kurt]), axis=1)
    gas_moments = np.zeros((4, 2*num_wells))
    gas_moments[0, num_wells:] = np.ones(num_wells)*50000
    _run_synthetic_case(oil_moments, gas_moments, 2)


def test_synthetic_case3():
    num_wells = 5
    mean = np.ones(num_wells)*50000.
    variance = np.arange(4000, 14000, 2000, dtype=np.float64)**2
    skew = np.zeros(num_wells)
    kurt = np.ones(num_wells)*3.*variance**2
    gas_moments = np.append(np.zeros((4, num_wells)), np.array([mean, variance, skew, kurt]), axis=1)
    oil_moments = np.zeros((4, 2*num_wells))
    oil_moments[0, num_wells:] = np.ones(num_wells)*50
    _run_synthetic_case(oil_moments, gas_moments, 3)


def test_synthetic_case4():
    num_wells = 5
    mean = np.ones(num_wells)*50000
    variance = np.ones(num_wells)*(5000**2)
    skew = np.array([-0.5, -0.3, 0, 0.3, 0.5])*variance**1.5
    kurt = np.ones(num_wells)*3.*variance**2
    gas_moments = np.append(np.zeros((4, num_wells)), np.array([mean, variance, skew, kurt]), axis=1)
    oil_moments = np.zeros((4, 2*num_wells))
    oil_moments[0, num_wells:] = np.ones(num_wells)*50
    _run_synthetic_case(oil_moments, gas_moments, 4)
